/*
 MIDI Sequencer
 
 Copyright (C) 2013 Vance Gloster

  Permission is hereby granted, free of charge, to any person obtaining a 
  copy of this software and associated documentation files (the "Software"), 
  to deal in the Software without restriction, including without limitation 
  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
  and/or sell copies of the Software, and to permit persons to whom the 
  Software is furnished to do so, subject to the following conditions:
  
  The above copyright notice and this permission notice shall be included in 
  all copies or substantial portions of the Software.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
 
 This sketch provides a fully functional MIDI sequencer, and is designed to 
 work with the SparkFun MIDI Shield.  The shield has MIDI in, MIDI out,
 three buttons and two potentiometers (knobs you can turn).
 
 If you hold the shield with the MIDI connectors at the bottom, the buttons,
 from left to right are Play, Stop and Record.  The knobs, again from left
 to right are tempo and velocity.
 
 The sketch comes up with the sequence from Pink Floyd's song On The Run
 (from the Dark Side of the Moon CD), which is legal "fair use" because I 
 am providing it for educational purposes only.  It's only 8 notes, which
 are listed on the Wikipedia page for the song, so I'm not revealing any
 secret musical arrangement.
 
 The sequencer plays a sequence of MIDI notes when you connect the MIDI 
 out of the shield to the MIDI in of a synthesizer and press the Play 
 button.  All the notes have an equal duration.  The green LED on the
 shield will light while it is playing.
 
 You can input a different series of MIDI notes to the sequencer to play 
 by connecting the MIDI out of a keyboard (or other MIDI instrument) to
 the MIDI in of the shield.  Note: There is a switch on the MIDI Shield, 
 and for inputting notes it must be in the Run position; for writing a 
 sketch to the Arduino it must be in the Prog position.
 
 To change the sequence, make sure the playback is stopped, press the 
 Record button, and the red LED will light.  Then play the notes you 
 want in the order you want them played.  When inputting, you do not need 
 to play fast or even, just in the right order.  If you make a mistake press 
 the Stop button and start over.  If you complete playing the sequence in 
 the correct order, press the Record button again.  it will immediately 
 begin playing your new sequence.

 
 The circuit for the MIDI shield can be found here:
 https://www.sparkfun.com/datasheets/DevTools/Arduino/MIDI_Shield-v13.pdf

 The sketch uses a timer library created by Simon Monk and Jack Christensen.
 You can find it here:
 https://github.com/JChristensen/Timer
 
 There is an out-of date page about it on the Arduino website here:
 http://playground.arduino.cc/Code/Timer
 
 Follow the instructions at the github site about unzipping the library in your
 Arduino library folder, and if necessary, from inside this project, select the 
 menu Sketch->Import Library and import the timer library.
 
 */
 #include "Timer.h"
 
 ////////////////////////////
 // Global constants
 
 // Device pin numbers
 int startSwitch = 2;
 int stopSwitch = 3;
 int recordSwitch = 4;
 int greenLED = 6;
 int redLED = 7;
 
 // Global variables
 Timer t;                        // The timer is used to schedule MIDI events
 int tempo = 150;                // Tempo in Beats Per Minute
 unsigned long delayTime = 100;  // Delay between notes, computed from tempo
 int noteValue = 16;             // This is 4 for quarter notes, 8 for eighth notes, etc.
 // The sequence contains the MIDI note numbers used in the sequence
 // A zero says it is the end of the sequence
 int sequence[128] = { 40, 43, 45, 43, 50, 48, 50, 52, 0 };
 int sequencePointer = -1;        // The offset within the sequence of the next note to be played
 int sequenceBuffer[128];        // Buffer we record to
 int sequenceBufferPointer;      // Offset in the buffer where the next note will go
 int initialPot1;                // Initial value of the first knob (A1).  Used to see if it has been moved.
 int initialPot2;                // Initial value of the second knob (A0).
 int pot1;                       // The last-read value of the first knob
 int pot2;                       // The last-read value of the second knob
 int velocity;                   // The velocity computed from the second knob
 
 // Global flags
 boolean changeTiming = false;   // If true, the first knob has been moved
 boolean running = false;        // Says whether we are currently playing the sequence
 boolean recording = false;      // Says whether we are currently recording a new sequence
 boolean grabNext = false;       // True in recording mode when the last byte was Note On
                                 // so the next byte will be the note number
 boolean readVelocity = false;   // True in recording mode when the last byte was the note number
                                 // so the next byte will be velocity
 
/////////////////////////////////////
// Read the tempo, and compute
// the delay between notes
 void adjustDelay()
 {
   // Compute the delay time
   int divisor = noteValue / 4;
   long tempoMsec = 60000 / tempo;
   if(divisor < 1)
     delayTime = tempoMsec * (4 / noteValue);
   else
     delayTime = tempoMsec / divisor;
   // Set the timer to trigger the first note in 10ms
   t.after(delayTime, sendNote, 0);
 }
 
/////////////////////////////////////
// Read the first knob, compute the 
// tempo, and call the function
// that computes the delay
 void setNewTempo()
 {
   // This sets a range of tempos from about 40 BPM to about 240 BPM
   tempo = 40 + ((1023 - pot1) / 5);
   adjustDelay();
 }

///////////////////////////////////////////
// The Arduino runs this once at power-up
void setup() 
{
  // Initialize the pins that have pushbuttons on them as inputs
  pinMode(startSwitch, INPUT);
  pinMode(stopSwitch, INPUT);
  pinMode(recordSwitch, INPUT);
  // Cause the CPU to pull up the switch lines to +5 when the button is not pushed
  digitalWrite(startSwitch,HIGH);
  digitalWrite(stopSwitch,HIGH);
  digitalWrite(recordSwitch,HIGH);
  // Make the LED lines outputs
  pinMode(greenLED, OUTPUT);
  pinMode(redLED, OUTPUT);
  // Turn the LEDs off
  digitalWrite(greenLED, HIGH);
  digitalWrite(redLED, HIGH);
  
  //  Set MIDI baud rate:
  Serial.begin(31250);
  // Read the knobs
  initialPot1 = analogRead(A1);
  initialPot2 = analogRead(A0);
}

///////////////////////////////////////////////////
// The Arduino constantly calls this over and over
void loop() 
{
  int val;
  if(recording)  // If currently recording from the MIDI in
  {
    t.update();  // Let the timer see if it's time to send a note
    
    // Read a byte from the MIDI in
    val = Serial.read();
    if(val >= 0 && sequenceBufferPointer < 127)
    {
      // If this is a note number from a note on message, 
      // write it to the temporary sequence
      if(grabNext && val < 128)
      {
        sequenceBuffer[sequenceBufferPointer++] = val;
        grabNext = false;
        readVelocity = true;
      }
      // If the last note on message is actually a 
      // kind of note off message (zero velocity)
      // remove it from the temporary sequence
      else if(readVelocity && val < 128)
      {
        if(val == 0)
          sequenceBufferPointer--;
        readVelocity = false;
      }
      // If it's a note on message, set the flag to grab the next byte
      if(val == 0x80)
        grabNext = true;
    }

    // If the stop button is pressed, stop playing and turn off the LEDs
    val = digitalRead(stopSwitch);
    if(val == LOW)
    {
      recording = false;
      digitalWrite(greenLED, HIGH);
      digitalWrite(redLED, HIGH);
    }
    
    // If the record button is pressed, that means overwrite the sequence
    // and start playing it
    val = digitalRead(recordSwitch);
    if(val == LOW)
    {
      // Stop recording
      recording = false;
      digitalWrite(redLED, HIGH);
      
      // Copy the data
      for (int i = 0; i < sequenceBufferPointer; ++i)
        sequence[i] = sequenceBuffer[i];
      sequence[sequenceBufferPointer] = 0;
        
      // Start playing
      digitalWrite(greenLED, LOW);
      running = true;
      // Set up so there is no note off, and it starts on the first note of the sequence
      sequencePointer = -1;
      if(changeTiming)
        setNewTempo();
      else
        adjustDelay();
    }
  }
  else if(running)  // If currently playing a sequence to the MIDI out
  {
    // If the stop button is pressed, stop playing and turn off the LEDs
    val = digitalRead(stopSwitch);
    if(val == LOW)
    {
      running = false;
      digitalWrite(greenLED, HIGH);
      digitalWrite(redLED, HIGH);
    }
    
    // Read the knobs
    pot1 = analogRead(A1);
    t.update();  // Let the timer see if it's time to send a note
    pot2 = analogRead(A0);
    t.update();  // Let the timer see if it's time to send a note
    
    // Convert knob 2 into the velocity
    velocity = 127 - (pot2 / 10);
    t.update();  // Let the timer see if it's time to send a note
    
    // If the first knob had not previously been moved, check to see if it has been
    if(!changeTiming && (pot1 - initialPot1 > 10 || initialPot1 - pot1 > 10))
      changeTiming = true;
    t.update();  // Let the timer see if it's time to send a note
  }
  else  // If currently stopped (not playing or recording)
  {
    t.update();  // Let the timer see if it's time to send a note
    
    // If the start button is pressed, switch to running mode and turn on 
    // the green LED and start the timer
    val = digitalRead(startSwitch);
    if(val == LOW)
    {
      recording = false;
      running = true;
      // Set up so there is no note off, and it starts on the first note of the sequence
      sequencePointer = -1;
      if(changeTiming)
        setNewTempo();
      else
        adjustDelay();
      digitalWrite(greenLED, LOW);
    }
    else
    {
      // Otherwise, see if the record button is pressed
      val = digitalRead(recordSwitch);
      if(val == LOW)
      {
        // Setup the record buffer, and turn recording on
        sequenceBufferPointer = 0;
        grabNext = false;
        readVelocity = false;
        running = false;
        recording = true;
        
        // Turn on the red LED
        digitalWrite(greenLED, HIGH);
        digitalWrite(redLED, LOW);
        
        // Wait for record button to go up
        while(val == LOW)
        {
          t.update();  // Let the timer see if it's time to send a note
          val = digitalRead(recordSwitch);
        }
        // Get rid of any buffered data from MIDI in
        while(Serial.read() != -1)
          ;
        // Debounce the switch
        delay(20);
      }
    }
  }
}

/////////////////////////////////////////////////
// This function is called by the timer, and it
// sends one playback note to MIDI out
 void sendNote(void *ignored)
 {
   // If playback is still running, set up the timer to
   // call this function again for the next note
   if(running)
   {
     if(changeTiming)
       setNewTempo();
     else
       t.after(delayTime, sendNote, 0);
   }
   
   // Send the note off for the previous note
   if(sequencePointer > -1)
     noteOff(0, sequence[sequencePointer], velocity);
   
   // Move to the next note in the sequence
   ++sequencePointer;
   if(sequence[sequencePointer] < 1)
     sequencePointer = 0;
     
   // If playback is still running do the next note on
   if(running)
     noteOn(0, sequence[sequencePointer], velocity);
 }

/******************************
 *  MIDI transmit subroutines
 ******************************/

/////////////////////////////////////
//  Does a MIDI note on.
void noteOn(int midiChannel, int pitch, int velocity) 
{
  int command = 0x90 + (midiChannel & 0x0F);
  Serial.write(command);    // Note-on command, with channel imbedded
  Serial.write(pitch & 0x7F);
  Serial.write(velocity & 0x7F);
}

/////////////////////////////////////
//  Does a MIDI note off.
void noteOff(int midiChannel, int pitch, int velocity) 
{
  int command = 0x80 + (midiChannel & 0x0F);
  Serial.write(command);    // Note-off command, with channel imbedded
  Serial.write(pitch & 0x7F);
  Serial.write(velocity & 0x7F);
}

/////////////////////////////////////
//  Does a MIDI program change.
void programChange(int midiChannel, int program) 
{
  int command = 0xC0 + (midiChannel & 0x0F);
  Serial.write(command);    // Program change command, with channel imbedded
  Serial.write(program & 0x7F);
}

/////////////////////////////////////
//  Does a MIDI clock.  
//  Should send 24 per quarter note, 
//  equally spaced.
void midiClock() 
{
  Serial.write(0xF8);    // MIDI clock
}

/////////////////////////////////////
//  Does a MIDI start.
void midiStart() 
{
  Serial.write(0xFA);    // MIDI start
}

/////////////////////////////////////
//  Does a MIDI stop.
void midiStop() 
{
  Serial.write(0xFC);    // MIDI stop
}

